#ifndef _FLSHAREDMEMORY
#define _FLSHAREDMEMORY

/*
 *
 * This is part of the ANL/MCS/FL software infrastructure.
 * written by Ivan R. Judson
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <stdlib.h>

/** This class encapsulates System V shared memory operations. */

class FLSharedMemory 
{
  /** @name Methods */
  //@{
public:

  /// This constructor creates a new shared memory segment

  FLSharedMemory(key_t key, int size, 
		  int mode = 0666) 
    {
      _key = key;
      _size = size;
      _mode = mode;

      shm_get_id();
      
      shm_attach();
    }
  
  /// This constructor attaches to the specified shared memory segment, if none was specified it creates a new one
  FLSharedMemory(int id = 0)
    {
      _id = id;

      if(!_id)
	{
	  shm_get_id();
	}
      shm_attach();
    }

  /// The destructor detaches the shared memory segment

  ~FLSharedMemory(void)
    {
      if(shmctl(_id, IPC_RMID) < 0)
	{
	  perror("sharedMemory::DetachMemory: Could not remove shm");
	}
    }
  
  /// Returns the address of the shared memory segment

  void *GetAddress(void) {
    return _buffer;
  }

  /// Returns the ID of the shared memory segment

  int GetID(void) {
    return _id;
  }

private:

  /// ask the operating system for the shared segment ID

  void shm_get_id(void) {
#ifdef DEBUG
    fprintf(stderr, "Asked for <KEY %d> <SIZE %d> <MODE %d>\n",
	    _key, _size, _mode);
#endif
    if((_id = shmget(_key, _size, _mode | IPC_CREAT)) < 0) 
      perror("sharedMemory::MakeMemory: Unable to shmget memory.");
#ifdef DEBUG
    fprintf(stderr, "Got <ID %d> for <KEY %d> <SIZE %d> <MODE %d>\n",
	    _id, _key, _size, _mode);
#endif
  }

  /// ask the operating system to attach to the specified segment

  void shm_attach(void) {
    if((long) (_buffer = shmat(_id, (char *)0, 0)) < 0) 
      perror("sharedMemory::MakeMemory: Unable to shmat memory.");
#ifdef DEBUG
    fprintf(stderr, "Got <ADDR %x>\n", _buffer);
#endif
  } 
  //@}
  
  /** @name Data */
  //@{
  
  /// The shared memory key

  key_t _key;
  
  /// The size in bytes of the shared memory segment

  int _size;

  /// The operating system permissions associated with the segment

  int _mode;

  /// The operating system assigned ID for the segment

  int _id;

  /// The actual memory that is being shared

  void *_buffer;
  //@}
};

#endif

