/*
 * This is part of the ANL/MCS/FL CAVEav distribution.
 *     written by Ivan R. Judson
 *
 */

#include <CAVcam.h>

#ifdef __sgi
#include <unistd.h>
#include <strings.h>
#endif

inline  void swap(float &a, float &b)
{
  float temp = a;
  a = b;
  b = temp;
}

GLfloat CAVcam::reflectionMatrix_[4][4] = { 1,0,0,0, 0,-1,0,0, 0,0,1,0, 0,0,0,1 };

CAVcam::CAVcam(float ex, float ey, float ez, 
	       float cx, float cy, float cz,
	       float ux, float uy, float uz)
{
  reflectImage_ = 0;
  
  _red = 1.0; _green = 0.0; _blue = 0.0;
  _eyex = ex; _eyey = ey; _eyez = ez;
  _centerx = cx; _centery = cy; _centerz = cz;
  _upx = ux; _upy = uy; _upz = uz;

  _captureFrames = 0;

  _modFrames = 1;

  _focalLength = 70.0;
}

void CAVcam::setupOffscreen(void)
{
#ifdef GLXPIXMAP
  static int configuration[] = {
    GLX_DOUBLEBUFFER, GLX_RGBA, GLX_DEPTH_SIZE, 16, GLX_RED_SIZE, 1,
    GLX_GREEN_SIZE, 1, GLX_BLUE_SIZE, 1, None
  };

#endif
#ifdef PBUFFER
  static int pbufconfig[] = {
    GLX_DRAWABLE_TYPE_SGIX, GLX_PBUFFER_BIT_SGIX, 
    GLX_RENDER_TYPE_SGIX, GLX_RGBA_BIT_SGIX, None
  };

  static int pattrib[] = {
    GLX_LARGEST_PBUFFER_SGIX, GLX_PRESERVED_CONTENTS_SGIX, GL_TRUE, None
  };

  GLXFBConfigSGIX *fbconfig;
  unsigned int pw, ph;
  int ni;
#endif

#ifdef GLXPIXMAP
  _camDisplay = CAVEXDisplay();
#endif
#ifdef PBUFFER
  _camDisplay = XOpenDisplay(NULL);
#endif

  if(_camDisplay == NULL)
    fprintf(stdout, "Couldn't get Display...\n");
  
#ifdef GLXPIXMAP
  _camVisual = CAVEXVisualInfo();
#endif
  
#ifdef PBUFFER
  fprintf(stderr, "1\n");
  fbconfig = glXChooseFBConfigSGIX(_camDisplay, DefaultScreen(_camDisplay), 
				 pbufconfig, &ni);

  if(fbconfig == NULL)
    fprintf(stderr, "ah 2\n");
  else 
    fprintf(stderr, "yay 2\n");

  _pbuffer = glXCreateGLXPbufferSGIX(_camDisplay, *fbconfig, _fb_width, 
				     _fb_height, pattrib);

  fprintf(stderr, "3\n");
  //  glXQueryGLXPbufferSGIX(_camDisplay, _pbuffer, GLX_WIDTH_SGIX, &pw);

  fprintf(stderr, "4\n");
  //  glXQueryGLXPbufferSGIX(_camDisplay, _pbuffer, GLX_HEIGHT_SGIX, &ph);

  fprintf(stderr, "PBUFFER SIZE %d %d\n", pw, ph);

#endif
  if(_camVisual == NULL) {
   fprintf(stdout, "Couldn't find visual...\n");
  }

#ifdef GLXPIXMAP
  _camContext = glXCreateContext(_camDisplay, _camVisual, NULL, GL_FALSE);
#endif

#ifdef PBUFFER
  _camContext = glXCreateContextWithConfigSGIX(_camDisplay, *fbconfig, 
					       GLX_RGBA_TYPE_SGIX, NULL, 
					       GL_TRUE);
#endif

  if(_camContext == NULL)
    fprintf(stdout, "Couldn't make context...\n");

#ifdef GLXPIXMAP
  pmap = XCreatePixmap(_camDisplay, RootWindow(_camDisplay, 
					       _camVisual->screen), _fg_width,
		       _fb_height, _camVisual->depth);
  glxpmap = glXCreateGLXPixmap(_camDisplay, _camVisual, pmap);
#endif
}

CAVcam::~CAVcam(void)
{
}

void CAVcam::Init(void)
{
}

void CAVcam::Update(void)
{
  float **tmat = new (float (*[4]));
  for(int i = 0; i < 4; i++)
    tmat[i] = new float[4];

  if( CAVEEye == CAVE_LEFT_EYE && CAVEMasterDisplay() &&
      !(_fc % _modFrames) && _captureFrames ) 
    {
      // Call the update

      if(UpdateFunc != NULL)
	{
	  (*UpdateFunc)();

	  // Generate the new position/orientation matrix
	  glPushMatrix();
	  {
	    glLoadIdentity();
	    gluLookAt(_eyex, _eyey, _eyez, 
		      _eyex + _centerx, _eyey + _centery, _eyez + _centerz, 
		      _upx, _upy, _upz);
	    glGetFloatv(GL_MODELVIEW_MATRIX, _invViewMat);
	  }
	  glPopMatrix();

#ifdef DEBUG
	  fprintf(stderr, "View Matrix:\n");
	  for(i = 0; i < 16; i+=4)
	    fprintf(stderr, " %f %f %f %f\n", _invViewMat[i], _invViewMat[i+1],
		    _invViewMat[i+2], _invViewMat[i+3]);
	  fprintf(stderr, "\n");
#endif

	  OpenGLToMatrix(_invViewMat, tmat);
	  ComputeInverse4(tmat);
	  MatrixToOpenGL(tmat, _invViewMat);

#ifdef DEBUG
	  fprintf(stderr, "Inverse View Matrix:\n");
	  for(i = 0; i < 16; i+=4)
	    fprintf(stderr, " %f %f %f %f\n", _invViewMat[i], _invViewMat[i+1],
		    _invViewMat[i+2], _invViewMat[i+3]);
	  fprintf(stderr, "\n");
#endif
	}
    }
}

void CAVcam::Draw3DGrid(void)
{
  float x, y, z;

  glPushMatrix();
  {
    glBegin(GL_LINES);
    {
      for(x = -5.0; x < 5.0; x += 1.0)
	for(y = 0.0; y < 10.0; y += 1.0)
	  for(z = -5.0; z < 5.0; z += 1.0)
	    {
	      glColor3f((x+5.0)/10.0, y/10.0, (z+5.0)/10.0);
	    
	      glVertex3f(5.0, y, z);
	      glVertex3f(-5.0, y, z);
	      
	      glVertex3f(x, 0.0, z);
	      glVertex3f(x, 10.0, z);
	      
	      glVertex3f(x, y, 5.0);
	      glVertex3f(x, y, -5.0);
	    }
    }
    glEnd();
  }
  glPopMatrix();
}

void CAVcam::Draw(void)
{
#ifndef DEBUG
  GLfloat cful[] = { -0.25, 0.25, -0.25 };
  GLfloat cfur[] = {  0.25, 0.25, -0.25 };
  GLfloat cfll[] = { -0.25, -0.25, -0.25 };
  GLfloat cflr[] = {  0.25, -0.25, -0.25 };

  GLfloat cbul[] = { -0.45, 0.45, 0.25 };
  GLfloat cbur[] = {  0.45, 0.45, 0.25 };
  GLfloat cbll[] = { -0.45, -0.45, 0.25 };
  GLfloat cblr[] = {  0.45, -0.45, 0.25 };
#endif

#ifdef DEBUG
  Draw3DGrid();
#endif

#ifdef VIEWFINDER
  if(viewFinder_ != NULL)
    viewFinder_->draw();
#endif

  glPushMatrix();
  {
    glMultMatrixf(_invViewMat);
    glScalef(1.8, 1.1, 1.0);

#ifdef DEBUG
    glBegin(GL_LINES);
    {
      glColor3f(0.0, 1.0, 0.0);
      
      glVertex3f(0.0, 0.0, 0.0);
      glVertex3f(0.0, 0.0, -1.0);
      
      glColor3f(1.0, 0.0, 0.0);
      
      glVertex3f(0.0, 0.0, 0.0);
      glVertex3f(0.0, 1.0, 0.0);
    }
    glEnd();
#endif
    
#ifndef DEBUG
    glColor3f(_red, _green, _blue);
    
    glBegin(GL_QUAD_STRIP);
    {
      glVertex3fv(cbul);
      glVertex3fv(cful);
      
      glVertex3fv(cbur);
      glVertex3fv(cfur);
      
      glVertex3fv(cblr);
      glVertex3fv(cflr);
      
      glVertex3fv(cbll);
      glVertex3fv(cfll);
      
      glVertex3fv(cbul);
      glVertex3fv(cful);
    }
    glEnd();
#endif

  }
  glPopMatrix();

}

void CAVcam::Process(void)
{
  GLint buff_id[1];
  FLImageDescriptor *image;
  int width, height, depth;
  int nextIndex;
  GLint bpp[1];

  if (CAVEEye != CAVE_LEFT_EYE || !CAVEMasterDisplay())
      return;

  double thisFrame = CAVEGetTime();
  double interFrame = 1.0 / (double) _modFrames;
  double delay = thisFrame - lastFrameTime_;

//  printf("CAVcam delay: %f interFrame: %f\n", delay, interFrame);

  if (delay < interFrame)
      return;

  lastFrameTime_ = thisFrame;

//  if( CAVEEye == CAVE_LEFT_EYE && CAVEMasterDisplay() &&
//      !(_fc % _modFrames) && _captureFrames)
    {
#ifdef GLXPIXMAP
      if(glXMakeCurrent(_camDisplay, glxpmap, _camContext))
	{
#endif
#ifdef PBUFFER
      if(glXMakeCurrent(_camDisplay, _pbuffer, _camContext))
        {
#endif
	  // Get the right place in the ring buffers to put the newest image
	  nextIndex = (_imagepool->_nextImage + 1) % _imagepool->_nImages;
	  image = &(_imagepool->_images[nextIndex]);

	  // Get these once
	  width = image->GetWidth();
	  height = image->GetHeight();
	  depth = image->GetDepth();

	  // Get which buffer we'll be reading from
	  glGetIntegerv(GL_AUX_BUFFERS, bpp);
	  glGetIntegerv(GL_DRAW_BUFFER, buff_id);

	  // Get window info
	  glPushAttrib(GL_VIEWPORT_BIT);

	  glMatrixMode(GL_PROJECTION);
	  glPushMatrix();
	  {
	    glLoadIdentity();

	
#ifdef DEBUG
	    fprintf(stderr, "done.\n");
	    fprintf(stderr, "Camera Position: < %f %f %f >\n", 
		    _eyex, _eyey, _eyez);
	    fprintf(stderr, "Camera Forward: < %f %f %f >\n", 
		    _eyex + _centerx, _eyey + _centery, _eyez + _centerz);
	    fprintf(stderr, "Camera Up: < %f %f %f >\n", _upx, _upy, _upz);
#endif

	    gluPerspective(_focalLength, (float)width/(float)height, 
			   CAVENear, CAVEFar);
	    glViewport(0, 0, width, height);

	    glMatrixMode(GL_MODELVIEW);
	    glPushMatrix();
	    {
	      glLoadIdentity();

	      if (reflectImage_)
	      {
		  glMultMatrixf((const GLfloat *)reflectionMatrix_);
	      }

	      // Set Camera
	      gluLookAt(_eyex, _eyey, _eyez, 
			_eyex + _centerx, _eyey + _centery, _eyez + _centerz, 
			_upx, _upy, _upz);

	      // Draw the scene
	      CallDraw();

	      // Wait for all graphics to get drawn
	      glFinish();

#if 0
	      _imagepool->_tmpim.SetSize(width, height, depth);

	      // Read Pixels
	      glReadPixels(0, 0, width, height, _format, GL_UNSIGNED_BYTE, 
			   (GLvoid *)_imagepool->_tmpim.GetData());
#elif 1
	      glReadPixels(0, 0, width, height, _format, GL_UNSIGNED_BYTE, 
			   (GLvoid *) image->GetData());
#else
	      glReadPixels(0, height, width, -height, _format, GL_UNSIGNED_BYTE, 
			   (GLvoid *) image->GetData());
#endif
	    }
	    glPopMatrix();
	    glMatrixMode(GL_PROJECTION);
	  }
	  glPopMatrix();
	  glMatrixMode(GL_MODELVIEW);
      
	  glPopAttrib();

	  // Update ring buffers
	  _imagepool->_nextImage = nextIndex;
	  _imagepool->_currentSerial++;
    
	  // Put them somewhere
	  PutPix();

	  Update();
#if defined(GLXPIXMAP) || defined(PBUFFER)
	}
      glXMakeCurrent(CAVEXDisplay(), CAVEXWindow(), CAVEGLXContext());
#endif
    }
}

FLImageDescriptor *CAVcam::getCurrentFrame()
{
  return &(_imagepool->_tmpim);
}

void CAVcam::PutPix(void)
{
}

void CAVcam::FlipFrame(int Index)
{
  FLImageDescriptor *image;
  volatile unsigned char *dataTo;
  volatile unsigned char *dataFrom;
  int i, stride, height, width, depth;

  // Set all pointers to point at the right things
  dataFrom = _imagepool->_tmpim.GetData();

  image = &(_imagepool->_images[Index]);
  dataTo = _imagepool->_images[Index].GetData();

  // retrieve these values once
  height = image->GetHeight();
  width = image->GetWidth();
  depth = image->GetDepth();

  // Calculate the size of one scanline
  stride = width * depth;

#ifdef DEBUG
  fprintf(stderr, "Size < %d %d %d %d >\n", height, width, depth, stride);
#endif
  
  // Loop through and copy each scanline to the correct line in the 
  // output frame buffer
  for(i = 0; i < height; i++)
    bcopy((void *)&(dataFrom[i * stride]), 
	  (void *)&(dataTo[(height - i) * stride]), 
	  stride);
}

int CAVcam::ComputeInverse4(float **a)
{
  // computes the inverse inplace
  // Gauss Jordan elimination from Numerical Recipes in C
  int indxc[4];
  int indxr[4];
  int ipiv[4];
  // size of the matrix
  const int n = 4;
  int i, icol, irow, j, k, l, ll;
  float big, dum, pivinv;
    
  for (j = 0; j < n; j++) ipiv[j] = 0;
  for (i = 0; i < n; i++)
    {
      big = 0.0;
      for (j = 0; j < n; j++)
	if (ipiv[j] != 1)
	  for (k = 0; k < n; k++)
	    {
	      if (ipiv[k] == 0)
		{
		  if (fabs(a[j][k]) >= big)
		    {
		      big = fabs(a[j][k]);
		      irow = j;
		      icol = k;
		    }
		}
	      else if (ipiv[k] > 1)
		{
		  fprintf(stderr, "Error: Singular Matrix-1");
		  return(0);
		}
	    }
      ++(ipiv[icol]);
	
      // interchange rows, if needed to get pivot element on diagonal
      if (irow != icol)
	{
	  for (l = 0; l < n; l++) swap(a[irow][l], a[icol][l]);
	}
      indxr[i] = irow;
      indxc[i] = icol;
      if (a[icol][icol] == 0.0)
	{
	  fprintf(stderr, "Error: Singular Matrix-2");
	  return(0);
	}
      pivinv = 1.0/a[icol][icol];
      a[icol][icol] = 1.0;
      for (l = 0; l < n; l++) a[icol][l] *= pivinv;
      for (ll = 0; ll < n; ll++)
	if (ll != icol)
	  {
	    dum = a[ll][icol];
	    a[ll][icol] = 0.0;
	    for (l = 0; l < n; l++) a[ll][l] -= a[icol][l] * dum;
	  }
    }
  // unscramble the solution
  for (l = n-1; l >=0; l--)
    {
      if (indxr[l] != indxc[l])
	for (k = 0; k < n; k++) swap(a[k][indxr[l]], a[k][indxc[l]]);
    }
  return(1);
}


