#include "TexturedPlane.h"

#include <GL/glu.h>
#include "cave_ogl.h"

#define TEX_EXT

TexturedPlane::TexturedPlane() : format_(GL_RGB)
{
    int i;
    
    textureSerial_ = 0;
    for (i = 0; i < CAVE_MAX_WALLS; i++)
    {
	firstBuild_[i] = 1;
	wallSerial_[i] = 0;
    }

    lock_ = CAVENewLock();

    textureWidth_ = textureHeight_ = 0;
    textureData_ = 0;
}

TexturedPlane::TexturedPlane(int width, int height, int depth)
{
    int i;
    
    textureSerial_ = 0;
    for (i = 0; i < CAVE_MAX_WALLS; i++)
    {
	firstBuild_[i] = 1;
	wallSerial_[i] = 0;
    }

    textureWidth_ = width;
    textureHeight_ = height;
    textureDepth_ = depth;

    nextpowtwowidth_ = FLUtils::getNextPowTwo(width);
    nextpowtwoheight_ = FLUtils::getNextPowTwo(height);

    lock_ = CAVENewLock();

    textureData_ = 0;
}

TexturedPlane::~TexturedPlane(void)
{

}

void setFormat(int format)
{

}

void TexturedPlane::setTexture(int width, int height, int depth,
			       unsigned char *data, GLenum format)
{
  CAVESetWriteLock(lock_);
  textureWidth_ = width;
  textureHeight_ = height;
  textureDepth_ = depth;
  textureData_ = data;
  format_ = format;
  int newP2w = FLUtils::getNextPowTwo(width);
  int newP2h = FLUtils::getNextPowTwo(height);
  if(newP2w != nextpowtwowidth_ || newP2h != nextpowtwoheight_)
    {
      nextpowtwowidth_ = FLUtils::getNextPowTwo(width);
      nextpowtwoheight_ = FLUtils::getNextPowTwo(height);
      for(int i = 0; i < CAVE_MAX_WALLS; i++)
	firstBuild_[i] = 1;
    }
#ifdef DEBUG
  fprintf(stderr, "OX: %d OY: %d NX: %d NY: %d\n", textureWidth_,
	  textureHeight_, nextpowtwowidth_, nextpowtwoheight_);
#endif
  textureSerial_++;
  CAVEUnsetWriteLock(lock_);
}

void TexturedPlane::appInit()
{
    int myWall = getWallIndex();

    dlIndex_[myWall] = glGenLists(1);

    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}

void TexturedPlane::frame()
{
    int myWall = getWallIndex();

    CAVESetReadLock(lock_);

    if (textureSerial_ > wallSerial_[myWall])
    {
	if (firstBuild_[myWall])
	{
	    buildWall(myWall);
	    firstBuild_[myWall] = 0;
	}
	else
	    rebuildWall(myWall);
    }

    CAVEUnsetReadLock(lock_);
}


#ifdef TEX_EXT
void TexturedPlane::buildWall(int myWall)
{
  glTexImage2D(GL_TEXTURE_2D, 0, 4, nextpowtwowidth_, nextpowtwoheight_,
	       0, format_, GL_UNSIGNED_BYTE, (GLvoid *) NULL);
  
  
  rebuildWall(myWall);
}

void TexturedPlane::rebuildWall(int myWall)
{
  float smin = 0.0, tmin = 0.0;
  float smax, tmax;

  smax = (float)textureWidth_ / (float)nextpowtwowidth_;
  tmax = (float)textureHeight_ / (float)nextpowtwoheight_;

  glTexSubImage2DEXT(GL_TEXTURE_2D, 0, 0, 0,
		     nextpowtwowidth_, nextpowtwoheight_, format_,
		     GL_UNSIGNED_BYTE, (GLvoid *)textureData_);
  
  glEnable(GL_TEXTURE_2D);

  glNewList(dlIndex_[myWall], GL_COMPILE);
  
  glBegin(GL_QUADS);
  {
    glTexCoord2f(smin, tmin);
    glVertex2f(-0.5, -0.5);
    glTexCoord2f(smin, tmax);
    glVertex2f(-0.5, 0.5);
    glTexCoord2f(smax, tmax);
    glVertex2f(0.5, 0.5);
    glTexCoord2f(smax, tmin);
    glVertex2f(0.5, -0.5);
  }
  glEnd();

  glEndList();
  
  glDisable(GL_TEXTURE_2D);

  wallSerial_[myWall] = textureSerial_;
}

#else /* TEX_EXT */

void TexturedPlane::buildWall(int myWall)
{
  float smin = 0.0, tmin = 0.0;
  float smax = (float)textureWidth_ / (float)nextpowtwowidth_;
  float tmax = (float)textureHeight_ / (float)nextpowtwoheight_;

  glTexImage2D(GL_TEXTURE_2D, 0, 3, nextpowtwowidth_, nextpowtwoheight_,
	       0, format_, GL_UNSIGNED_BYTE, (GLvoid *)textureData_);
  glEnable(GL_TEXTURE_2D);

  glNewList(dlIndex_[myWall], GL_COMPILE);

  glBegin(GL_QUADS);
  {
    glTexCoord2f(smin, tmin);
    glVertex2f(-0.5, -0.5);
    glTexCoord2f(smin, tmax);
    glVertex2f(-0.5, 0.5);
    glTexCoord2f(smax, tmax);
    glVertex2f(0.5, 0.5);
    glTexCoord2f(smax, tmin);
    glVertex2f(0.5, -0.5);
  }
  glEnd();

  glEndList();

  glDisable(GL_TEXTURE_2D);
  
  wallSerial_[myWall] = textureSerial_;
}

void TexturedPlane::rebuildWall(int myWall)
{
    buildWall(myWall);
}

#endif /* TEX_EXT */

void TexturedPlane::draw()
{
    glPushMatrix();
    {
      glEnable(GL_COLOR_MATERIAL);
      glEnable(GL_TEXTURE_2D);

      glCallList(dlIndex_[getWallIndex()]);

      glDisable(GL_TEXTURE_2D);
      glDisable(GL_COLOR_MATERIAL);
    }
    glPopMatrix();
}

int TexturedPlane::writePaneEPS(char *filename)
{
  FILE *fp;
  int components, pos, i;
  char *curpix;

  if(format_ == GL_RGB)
    components = 3;
  else
    components = 4;

  if((fp = fopen(filename, "w")) == NULL)
    return 1;
  else
    {
      fprintf(fp, "%%!PS-Adobe-2.0 EPSF-1.2\n");
      fprintf(fp, "%%%%Creator: CAVEav 2.0 CAVE output\n");
      fprintf(fp, "%%%%BoundingBox: 0 0 %d %d\n", 
	      textureWidth_, textureHeight_);
      fprintf(fp, "%%%%EndComments\n");
      fprintf(fp, "gsave\n");
      fprintf(fp, "/bwproc {\n");
      fprintf(fp, "    rgbproc\n");
      fprintf(fp, "    dup length 3 idiv string 0 3 0\n");
      fprintf(fp, "    5 -1 roll {\n");
      fprintf(fp, "    add 2 1 roll 1 sub dup 0 eq\n");
      fprintf(fp, "    { pop 3 idiv 3 -1 roll dup 4 -1 roll dup\n");
      fprintf(fp, "       3 1 roll 5 -1 roll } put 1 add 3 0 \n");
      fprintf(fp, "    { 2 1 roll } ifelse\n");
      fprintf(fp, "    } forall\n");
      fprintf(fp, "    pop pop pop\n");
      fprintf(fp, "} def\n");
      fprintf(fp, "systemdict /colorimage know not {\n");
      fprintf(fp, "   /colorimage {\n");
      fprintf(fp, "       pop\n");
      fprintf(fp, "       pop\n");
      fprintf(fp, "       /rgbproc exch def\n");
      fprintf(fp, "       { bwproc } image\n");
      fprintf(fp, "   } def\n");
      fprintf(fp, "} if\n");
      fprintf(fp, "picstr %d string def\n", textureWidth_ * components);
      fprintf(fp, "%d %d scale\n", textureWidth_, textureHeight_);
      fprintf(fp, "%d %d %d\n", textureWidth_, textureHeight_, 8);
      fprintf(fp, "[%d 0 0 %d 0 0]\n", textureWidth_, textureHeight_);
      fprintf(fp, "{ currentfile picstr readhexstring pop }\n");
      fprintf(fp, "false %d\n", components);
      fprintf(fp, "colorimage\n");

      curpix = (char *)textureData_;
      pos = 0;
      for(i = textureWidth_ * textureHeight_ * components; i > 0; i--) {
	fprintf(fp, "%02hx", *curpix++);
	if(++pos >= 32) {
	  fprintf(fp, "\n");
	  pos = 0;
	}
      }

      if(pos)
	fprintf(fp, "\n");
      
      fprintf(fp, "grestore\n");
      fclose(fp);
      return(0);
    }
}
